(** example TAL program: chase *)

(** Chunxiao Lin @ 2006 *)

Set Implicit Arguments.
Unset Strict Implicit.

Require Export talrules.


(** ******************************************************************* *)
(** list type *)
(** ******************************************************************* *)
Notation tlist := (trec (tunion tnull (tmem true tint true (tvar 0)))).


(** ******************************************************************* *)
(** code and sepcifications *)
(** ******************************************************************* *)

Notation INIT := 25.
Notation CHASE := 26.
Notation WRITE := 27.
Notation RET := 28.

Definition regty_cons' := (fun R G => regty_cons (fst R) (snd R) G).

Definition init_G := [{ (s1, tlist), (zero, tint) }].

Definition chase_G := (regty_set init_G s1 (rec_unfold tlist)).

Definition write_G := [{ (s1, tlist), 
  (s2, [< tint ^^ false, tlist ^^ false >]),
  (zero, tint) }].

Definition ret_G := [{}].

Definition alloc_G := [{ (s1, tlist),
  (zero, tint),
  (ra, (tcode 
    [{ (s1, tlist), (s2, [< tint ^^ false, tlist ^^ false >]), (zero, tint) }]))
  }].

Definition init := 
  j CHASE.

Definition chase := 
  beq s1 zero RET;;
  lw s1 4 s1;
  jal WRITE ALLOC.

Definition write :=
  sw zero 0 s2;
  sw s1 4 s2;
  addiu s1 s2 0;
  j CHASE.

Definition ret := 
  j RET.

Definition chase_P :=
  (INIT,,(isp1 init_G)):::(
  (CHASE,,(isp1 chase_G)):::(
  (WRITE,,(isp1 write_G)):::(
  (RET,,(isp1 ret_G)):::(
  (ALLOC,,(isp1 alloc_G)):::(
  nilT _))))).

Lemma WFregty_chase : WFregty chase_G.
Proof.
  split; trivial.
  split; trivial.
  intros r F; destruct F as [R | [R | R]]; subst; try tauto; intuition.
  split; trivial.
  simpl; intuition.
  discriminate.
  intuition.
Qed.

Lemma WFregty_write : WFregty write_G.
Proof.
  split; trivial.
  split; trivial.
  intros r F; destruct F as [R | [R | [R | R]]]; subst; try tauto; intuition.
  split; trivial.
  simpl; intuition; discriminate.
  intuition.
Qed.

Lemma WFregty_ret : WFregty ret_G.
Proof.
  split; trivial.
  split; trivial.
  simpl; intros r F; destruct F.
  split; trivial.
  simpl; intuition; discriminate.
  intuition.
Qed.

Lemma WFregty_init : WFregty init_G.
Proof.
  split; trivial.
  split; trivial.
  intros r F; destruct F as [R | [R | R]]; subst; try tauto; intuition.
  split; trivial.
  simpl; intuition; discriminate.
  intuition.
Qed.

Lemma WFregty_alloc : WFregty alloc_G.
Proof.
  split; trivial.
  split; trivial.
  intros r F; destruct F as [R | [R |[ R | R]]]; subst; try tauto; intuition.
  split; trivial.
  simpl; intuition; discriminate.
  intuition.
Qed.

Theorem init_ok :
  WFis chase_P init_G init. 
Proof.
  unfold init.
  eapply WFis_j.
  left; reflexivity.
  unfold chase_G.
  apply regty_sub_unfold; trivial.
  discriminate.
  apply WFregty_init.
Qed.
  
Theorem chase_ok :
  WFis chase_P chase_G chase. 
Proof.
  unfold chase.
  eapply WFis_nulle.
  simpl; reflexivity.
  left; reflexivity.
  constructor; simpl; intros; try tauto.
  apply WFregty_ret.

  eapply WFis_lw4.
  discriminate.
  intuition.
  simpl; reflexivity.

  eapply WFis_jal.
  left; reflexivity.
  left; reflexivity.
  apply WFregty_write.
  simpl.
  constructor.
  intros.
  destruct H as [R |[ R |[ R | R]]]; try tauto; subst; reflexivity.
  apply WFregty_alloc.
Qed.

Theorem write_ok :
  WFis chase_P write_G write. 
Proof.
  unfold write.
  eapply WFis_sw0.
  simpl; reflexivity.
  simpl; reflexivity.

  eapply WFis_sw4.
  simpl; reflexivity.
  simpl; reflexivity.

  eapply WFis_mov.
  discriminate.
  intuition.
  simpl; reflexivity.
  instantiate (1 := tunion tnull (tmem true tint true tlist)).
  constructor.
  unfold WFwty; reflexivity.

  eapply WFis_j.
  left; reflexivity.
  constructor;intros.
  destruct H as [R | [R | R]]; try tauto; subst; simpl; reflexivity.
  apply WFregty_chase.
Qed.

Theorem ret_ok :
  WFis chase_P ret_G ret. 
Proof.
  unfold ret.
  eapply WFis_j.
  left; reflexivity.
  constructor; intros.
  destruct H.
  apply WFregty_ret.
Qed.