(** TAL on LOCAP *)

(** Typing rules & Soundness *)

(** Chunxiao Lin @ 2006 *)

Set Implicit Arguments.
Unset Strict Implicit.

Require Export locap.


(** ******************************************************************* *)
(** lemmas *)
(** ******************************************************************* *)

(** register set *)
Lemma WFregty_rset : forall G r T,
  r <> zero -> rlst_in r tal_regs -> WFregty G -> 
  WFwty T -> WFregty (regty_set G r T).
Proof.
  intros G r T Hz Hin [Hb [Hd [Hc Hzero]]] Ht.
  assert (Hc' := (regty_canon_rset r T Hc)).
  repeat (split; trivial).
  apply rec_bound_regty_intro; auto.
  intros.
  comparer r r0.
  rewrite regty_set_get_eq in H.
  injection H; intros; subst.
  apply Ht.
  rewrite regty_set_get_neq in H; auto.
  apply (rec_bound_regty_elim Hc Hb H).
  red; intros.
  comparer r0 r.
  trivial.
  apply Hd.
  apply regty_set_dom_subset2 with r T; auto.
  rewrite regty_set_get_neq; firstorder.
Qed.

Lemma WFrf_rset : forall Psi Phi G rf r w T,
  r <> zero ->
  rlst_in r tal_regs ->
  WFrf Psi Phi rf G -> 
  WFw Psi Phi w T ->
  WFwty T ->
  WFrf Psi Phi (rfile_set rf r w) (regty_set G r T).
Proof.
  split.
  intros.
  comparer r0 r; solvereg.
  rewrite rfile_set_get_eq; auto.
  rewrite regty_set_get_eq in H4.
  injection H4; intros; subst.
  trivial.
  rewrite (regty_set_get_neq G T e) in H4.
  destruct H1.
  rewrite rfile_set_get_neq; auto.
  destruct H1.
  apply WFregty_rset; auto.
Qed.

Lemma WFregty_set : forall G r T,
  r <> zero ->
  WFwty T -> 
  regty_indom r G ->
  WFregty G ->
  WFregty (regty_set G r T).
Proof.
  intros G r T Hz HT Hdom [Hr [Hb [Hc Hzero]]].
  assert (Hc' := regty_canon_rset r T Hc).
  repeat (split; trivial).
  apply rec_bound_regty_intro; trivial.
  intros.
  hypreplace Hr (rec_bound_regty_elim Hc Hr).
  comparer r0 r.
  rewrite regty_set_get_eq in H.
  injection H; intros; subst.
  apply HT.
  rewrite regty_set_get_neq in H; firstorder.
  red; intros.
  comparer r0 r; apply Hb.
  apply Hdom.
  apply regty_set_dom_subset2 with r T; trivial.
  rewrite regty_set_get_neq; firstorder.
Qed.

(** WFwty intro *)
Lemma regty_bound_WFwty : forall G r T,
  rec_bound_regty 0 G = true -> 
  regty_get G r = Some T ->
  WFwty T.
Proof.
  induction G; simpl.
  intros; discriminate.
  intros r' T Hb.
  elimbool.
  comparer r' r; solvereg; intros.
  injection H; intros; subst.
  apply B.
  firstorder.
Qed.

Lemma WFwty_wty_sub : forall T T',
  WFwty T -> wty_sub T T' -> WFwty T'.
Proof.
  intros T T' Hwf Hsub.
  induction Hsub; trivial;
    generalize Hwf; try (generalize H); clear Hwf; try clear H;
      unfold WFwty.
  simpl; intros; splitb; trivial.
  simpl; intros; splitb; trivial.
Qed.

(** rfile subtype ok *)
Lemma WFst_regty_sub : forall Psi G G' st,
  WFst Psi G st -> regty_sub G G' -> WFst Psi G' st.
Proof.
  intros Psi G G' [hp rf] [Phi [Hdh [Hrf WFrty]]] Hsub.
  exists Phi.
  split; trivial.
  split; trivial.
  intros.
  induction Hsub.

  apply Hrf.
  rewrite <- (H0 _ (regty_indom_intro H)).
  apply H.
  
  comparer r0 r.
  rewrite regty_set_get_eq in H.
  hypreplace H1 (Hrf _ _ H1).
  injection H; intros; subst.
  constructor; trivial.
  rewrite regty_set_get_neq in H; auto.

  comparer r0 r.
  rewrite regty_set_get_eq in H.
  hypreplace H1 (Hrf _ _ H1).
  injection H; intros; subst; clear H.
  inversion H1. 
  apply H6.
  inversion H3.
  rewrite regty_set_get_neq in H; auto.

  induction Hsub; trivial.
  apply WFregty_rset; trivial.
  hypreplace H0 (regty_indom_intro H0).
  destruct H1 as [Hr [Hb Hc]].
  apply Hb.
  apply H0.
  
  destruct H1 as [Hr [Hb [Hc Hz]]].
  unfold WFwty.
  apply rec_bound_fold.
  hypreplace WFrty (regty_bound_WFwty Hr H0).
  apply WFrty.

  apply WFregty_rset; trivial.
  hypreplace H0 (regty_indom_intro H0).
  destruct H1 as [Hr [Hb Hc]].
  apply Hb.
  apply H0.
  
  destruct H1 as [Hr [Hb [Hc Hz]]].
  unfold WFwty.
  apply rec_bound_unfold.
  hypreplace WFrty (regty_bound_WFwty Hr H0).
  apply WFrty.
Qed.


(** cellty *)
Lemma cellty_sub_T_eq : forall b1 T1 b2 T2,
  cellty_sub (b1, T1) (b2, T2) -> T1 = T2.
Proof.
  intros.
  inversion H; trivial.
Qed.

Lemma cellty_sub_trans : forall C1 C2 C3,
  cellty_sub C1 C2 -> cellty_sub C2 C3 -> cellty_sub C1 C3.
Proof.
  intros.
  destruct C1; destruct C2; destruct C3.
  inversion H.
  destruct C.
  subst.
  trivial.
  subst.
  inversion H0; subst.
  trivial.
Qed.

Ltac injcty :=
  match goal with
    | [Hcsub : cellty_sub (?b1, ?T1) (?b2, ?T2) |- _] =>
      let e := fresh "e" in
	assert (e := cellty_sub_T_eq Hcsub); subst T1
  end.

(* Axiom fs : False. *)
(* Ltac skipgoal := elimtype False; apply fs. *)

(** word value canonical form *)
Lemma WFw_canon_tnull : forall Psi Phi w,
  WFw Psi Phi w tnull -> w = 0.
Proof.
  intros.
  inversion_clear H.
  reflexivity.
  inversion H1.
Qed.

Lemma WFw_canon_tvar : forall Psi Phi w n,
  WFw Psi Phi w (tvar n) -> False.
Proof.
  intros.
  inversion H.
  inversion H1.
Qed.

Lemma WFw_canon_tmem : forall Psi Phi w b1 T1 b2 T2,
  WFw Psi Phi w (tmem b1 T1 b2 T2) -> 
  exists b1', exists b2', exists T1', exists T2',
    Phi w = Some ((b1', T1'),(b2', T2')) /\
    cellty_sub (b1', T1') (b1, T1) /\
    cellty_sub (b2', T2') (b2, T2).
Proof.
  intros.
  inversion_clear H.
  destruct C1; destruct C2.
  exists i; exists i0; exists w0; exists w1.
  firstorder.
  inversion H1.
Qed.

Lemma WFw_canon_tunion : forall Psi Phi w T1 T2,
  WFw Psi Phi w (tunion T1 T2) -> 
  WFw Psi Phi w T1 \/ WFw Psi Phi w T2.
Proof.
  intros.
  inversion_clear H.
  inversion H1; subst; firstorder.
Qed.

Lemma WFw_canon_tunion2 : forall Psi Phi w T1 T2,
  WFw Psi Phi w T1 \/ WFw Psi Phi w T2 ->
  WFwty T1 -> WFwty T2 ->
  WFw Psi Phi w (tunion T1 T2).
Proof.
  intros.
  destruct H.
  apply WFw_sub with T1; trivial.
  constructor; trivial.
  apply WFw_sub with T2; trivial.
  constructor; trivial.
Qed.

(** WFwty *)
Lemma WFwty_tunion_elim : forall T1 T2,
  WFwty (tunion T1 T2) -> WFwty T1 /\ WFwty T2.
Proof.
  intros T1 T2.
  unfold WFwty; simpl.
  intro; elimbool; firstorder.
Qed.

(** heap extension *)
Notation dhspec_set := (fun Phi l (C1 C2 : cellty) => 
  (fun l' => if (beq_nat l' l) then Some (C1, C2) else Phi l')).

Lemma dhspec_set_WFdhspec : forall Phi l b1 T1 b2 T2,
  WFdhspec Phi ->
  WFwty T1 -> WFwty T2 ->
  vptr l ->
  WFdhspec (dhspec_set Phi l (b1, T1) (b2, T2)).
Proof.
  red; intros.
  generalize H3; clear H3.
  simplif; intros.
  injection H3; intros; subst; firstorder.
  apply (H _ _ _ _ _ H3).
Qed.

Lemma heap_extension_WFw : forall Tv Psi Phi l v C1 C2,
  WFw Psi Phi v Tv ->
  WFwty Tv ->
  Phi l = None ->
  WFw Psi (dhspec_set Phi l C1 C2) v Tv.
Proof.
  intros Tv Psi Phi l v C1 C2 Hv Htv Hl.
  induction Hv.
  
  constructor.
  constructor.
  apply WFw_code; trivial.
  compare l addr; intro e; subst.
  rewrite Hl in H; discriminate.
  apply WFw_mem with C0 C3; trivial; simplbeq; trivial.
  constructor.
  apply IHHv; trivial.
  apply (rec_bound_unfold Htv).
  induction H.
  constructor.
  destruct (WFwty_tunion_elim Htv).
  apply WFw_sub with T.
  apply IHHv; trivial.
  constructor; trivial.
  destruct (WFwty_tunion_elim Htv).
  apply WFw_sub with T'.
  apply IHHv; trivial.
  constructor; trivial.
Qed.

Lemma heap_extension_WFcell : forall bc Tc Psi Phi l v C1 C2,
  WFcell Psi Phi v (bc, Tc) ->
  WFwty Tc ->
  Phi l = None ->
  WFcell Psi (dhspec_set Phi l C1 C2) v (bc, Tc).
Proof.
  intros bc Tc Psi Phi l v C1 C2 Hv Htv Hl.
  induction bc.
  inversion Hv; subst.
  constructor.
  apply heap_extension_WFw; trivial.
  constructor.
Qed.

Lemma heap_extension_WFrf : forall G rf Psi Phi l C1 C2,
  WFrf Psi Phi rf G ->
  Phi l = None ->
  WFrf Psi (dhspec_set Phi l C1 C2) rf G.
Proof.
  intros G rf Psi Phi l C1 C2 [Hr Hwf] Hl.
  split; trivial.
  intros.
  hypreplace Hr (Hr _ _ H).
  apply heap_extension_WFw; trivial.
  apply regty_bound_WFwty with G r; trivial.
  destruct Hwf; firstorder.
Qed.

Notation WFcellty := (fun C => WFwty (snd C)).

Lemma heap_extension_WFdh : forall hp Psi Phi l v1 v2 C1 C2,
  WFdh Psi Phi hp Phi ->
  vptr l ->
  Phi l = None ->
  WFcell Psi Phi v1 C1 ->
  WFcell Psi Phi v2 C2 ->
  WFcellty C1 -> WFcellty C2 ->
  WFdh Psi (dhspec_set Phi l C1 C2) 
  (heap_set (heap_set hp l v1) (l + 4) v2) (dhspec_set Phi l C1 C2).
Proof.
  intros hp Psi Phi l v1 v2 C1 C2 [Hhp Hwf] [Hl1 Hl2] Hl3 Hv1 Hv2 Hcty1 Hcty2.
  split.

  intros l' C3 C4.
  simplif; subst; intros.
  injection H; clear H; intros; subst.
  exists v1; exists v2.
  rewrite heap_set_get_ne.
  repeat (rewrite heap_set_get_eq).
  repeat split; trivial.
  destruct C3; apply heap_extension_WFcell; trivial.
  destruct C4; apply heap_extension_WFcell; trivial.
  apply (mod8_plus4_ptr Hl2).
  apply (mod8_ptr Hl2).
  omega.
  
  destruct (Hhp _ _ _ H) as [v3 [v4 [Hv3 [Hv4 [Hvc1 Hvc2]]]]].
  exists v3; exists v4.
  destruct C3; destruct C4.
  destruct (Hwf _ _ _ _ _ H).
  destruct H0 as [_ A]. 
  repeat (rewrite heap_set_get_ne); trivial.
  repeat split; trivial.
  apply heap_extension_WFcell; intuition.
  apply heap_extension_WFcell; intuition.
  intro F; subst l; 
    rewrite (mod8_plus4 A) in Hl2; discriminate.
  omega.
  intro F; subst l'; 
    rewrite (mod8_plus4 Hl2) in A; discriminate.
  destruct C1; destruct C2.
  apply dhspec_set_WFdhspec; firstorder.
Qed.

(** heap update *)
Lemma heap_update_WFw : forall Tv Psi Phi l v bo bon To bt btn Tt,
  WFw Psi Phi v Tv ->
  WFwty Tv ->
  Phi l = Some ((bo, To), (bt, Tt)) ->
  cellty_sub (bon, To) (bo, To) ->
  cellty_sub (btn, Tt) (bt, Tt) ->
  WFw Psi (dhspec_set Phi l (bon, To) (btn, Tt)) v Tv.
Proof.
  intros Tv Psi Phi l v bo bon To bt btn Tt Hv Htv Hl Hc1 Hc2.
  induction Hv.
  
  constructor.
  constructor.
  apply WFw_code; trivial.

  destruct C1 as [b1' T1']; destruct C2 as [b2' T2'].
  repeat injcty.
  compare l addr; intro e; subst.
  rewrite H in Hl; injection Hl; clear Hl; intros; subst.
  apply WFw_mem with (bon,To) (btn,Tt); trivial.
  simplbeq; reflexivity.
  apply (cellty_sub_trans Hc1 H0).
  apply (cellty_sub_trans Hc2 H1).
  apply WFw_mem with (b1',T1) (b2',T2); trivial.
  simplbeq; trivial.

  constructor.
  apply IHHv; trivial.
  apply (rec_bound_unfold Htv).
  
  induction H.
  constructor.
  destruct (WFwty_tunion_elim Htv).
  apply WFw_sub with T.
  apply IHHv; trivial.
  constructor; trivial.
  destruct (WFwty_tunion_elim Htv).
  apply WFw_sub with T'.
  apply IHHv; trivial.
  constructor; trivial.
Qed.

Lemma  heap_update_WFcell : forall bc Tc Psi Phi l v bo bon To bt btn Tt,
  WFcell Psi Phi v (bc, Tc) ->
  WFwty Tc ->
  Phi l = Some ((bo, To), (bt, Tt)) ->
  cellty_sub (bon, To) (bo, To) ->
  cellty_sub (btn, Tt) (bt, Tt) ->
  WFcell Psi (dhspec_set Phi l (bon, To) (btn, Tt)) v (bc, Tc).
Proof.
  intros bc Tc Psi Phi l v bo bon To bt btn Tt Hv Htv Hl Hc1 Hc2.
  induction bc.
  inversion Hv; subst.
  constructor.
  apply heap_update_WFw with bo bt; trivial.
  constructor.
Qed.

Lemma heap_update_WFrf : forall G rf Psi Phi l bo bon To bt btn Tt,
  WFrf Psi Phi rf G ->
  Phi l = Some ((bo, To), (bt, Tt)) ->
  cellty_sub (bon, To) (bo, To) ->
  cellty_sub (btn, Tt) (bt, Tt) ->
  WFrf Psi (dhspec_set Phi l (bon, To) (btn, Tt)) rf G.
Proof.
  intros G rf Psi Phi l bo bon To bt btn Tt [Hr Hwf] Hl Hc1 Hc2.
  split; trivial.
  intros.
  hypreplace Hr (Hr _ _ H).
  apply heap_update_WFw with bo bt; trivial.
  apply regty_bound_WFwty with G r; trivial.
  destruct Hwf; firstorder.
Qed.

Lemma heap_update_WFdh_o : forall hp Psi Phi l v b bn T C,
  WFdh Psi Phi hp Phi ->
  Phi l = Some ((b, T), C) ->
  cellty_sub (bn, T) (b, T) ->
  WFw Psi Phi v T ->
  WFdh Psi (dhspec_set Phi l (bn, T) C) (heap_set hp l v) (dhspec_set Phi l (bn, T) C).
Proof.
  intros hp Psi Phi l v b bn T C [Hhp Hwf] Hl Hc Hv.
  split.

  intros l' [b' T'] C2.
  simplif; subst; intros.
  injection H; clear H; intros; subst.
  destruct (Hhp _ _ _ Hl) as [v1 [v2 [Hv1 [Hv2 [Hvc1 Hvc2]]]]].

  exists v; exists v2.
  assert (l <> l + 4); [omega | simplhp].
  destruct C2 as [b2 T2].
  repeat (split; trivial).
  destruct b'.
  constructor.
  apply heap_update_WFw with b b2; trivial.
  destruct (Hwf _ _ _ _ _ Hl); firstorder.
  constructor.
  constructor.
  apply heap_update_WFcell with b b2; trivial.
  destruct (Hwf _ _ _ _ _ Hl); firstorder.
  constructor.
  apply heap_get_ptr with hp v1; firstorder.
  destruct (Hhp _ _ _ H) as [v1 [v2 [Hv1 [Hv2 [Hvc1 Hvc2]]]]].
  exists v1; exists v2.
  assert (l <> l' + 4).
  destruct C;  destruct C2.
  destruct (Hwf _ _ _ _ _ Hl).
  destruct (Hwf _ _ _ _ _ H).
  intro F; subst.
  destruct H2; destruct H0.
  rewrite (mod8_plus4 H4) in H5; discriminate.
  simplhp.
  destruct C; destruct C2.
  destruct (Hwf _ _ _ _ _ H).
  repeat (split; trivial);
    apply heap_update_WFcell with b i; try constructor; firstorder.

  destruct C.
  destruct (Hwf _ _ _ _ _ Hl).
  apply dhspec_set_WFdhspec; firstorder.
Qed.

Lemma heap_update_WFdh_t : forall hp Psi Phi l v b bn T C,
  WFdh Psi Phi hp Phi ->
  Phi l = Some (C, (b, T)) ->
  cellty_sub (bn, T) (b, T) ->
  WFw Psi Phi v T ->
  WFdh Psi (dhspec_set Phi l C (bn, T)) (heap_set hp (l + WSIZE) v) (dhspec_set Phi l C (bn, T)).
Proof.
  intros hp Psi Phi l v b bn T C [Hhp Hwf] Hl Hc Hv.
  split.

  intros l' C1 [b' T'].
  simplif; subst; intros.
  injection H; clear H; intros; subst.
  destruct (Hhp _ _ _ Hl) as [v1 [v2 [Hv1 [Hv2 [Hvc1 Hvc2]]]]].

  exists v1; exists v.
  assert (l <> l + 4); [omega | simplhp].
  destruct C1 as [b1 T1].
  repeat (split; trivial).
  apply heap_update_WFcell with b1 b; trivial.
  destruct (Hwf _ _ _ _ _ Hl); firstorder.
  constructor.

  destruct b'.
  constructor.
  apply heap_update_WFw with b1 b; trivial.
  destruct (Hwf _ _ _ _ _ Hl); firstorder.
  constructor.
  constructor.
  apply heap_get_ptr with hp v2; firstorder.

  destruct (Hhp _ _ _ H) as [v1 [v2 [Hv1 [Hv2 [Hvc1 Hvc2]]]]].
  exists v1; exists v2.
  assert (l' <> l + 4).
  destruct C;  destruct C1.
  destruct (Hwf _ _ _ _ _ Hl).
  destruct (Hwf _ _ _ _ _ H).
  intro F; subst.
  destruct H0; destruct H2.
  rewrite (mod8_plus4 H4) in H5; discriminate.
  assert (l + 4 <> l' + 4);[omega | simplhp].
  destruct C; destruct C1.
  destruct (Hwf _ _ _ _ _ H).
  repeat (split; trivial);
    apply heap_update_WFcell with i b; try constructor; firstorder.

  destruct C.
  destruct (Hwf _ _ _ _ _ Hl).
  apply dhspec_set_WFdhspec; firstorder.
Qed.


(** ******************************************************************* *)
(** well-formed instruction *)
(** ******************************************************************* *)
Inductive WFis : chspec -> regty -> iseq -> Prop :=
  | WFis_j : 
    forall Phi G G' lab,
      chspec_in Phi lab (isp1 G') ->
      regty_sub G G' ->
      WFis Phi G (j lab)

  | WFis_jal : 
    forall Phi G G' G'' lab fret,
      chspec_in Phi lab (isp1 G') ->
      chspec_in Phi fret (isp1 G'') ->
      WFregty G'' ->
      regty_sub (regty_set G ra (tcode G'')) G' ->
      WFis Phi G (jal fret lab)

  | WFis_mov : 
    forall Phi G T T' rd rs is',
      (rd <> zero) ->
      (rlst_in rd tal_regs) ->
      regty_get G rs = Some T ->
      wty_sub T T' ->
      WFis Phi (regty_set G rd T') is' ->
      WFis Phi G (seq (addiu rd rs 0) is')

(*   | WFis_subt :  *)
(*     forall Phi G T T' rs is', *)
(*       (rs <> zero) -> *)
(*       (rlst_in rs tal_regs) -> *)
(*       regty_get G rs = Some T -> *)
(*       wty_sub T T' -> *)
(*       WFis Phi (regty_set G rs T') is' -> *)
(*       WFis Phi G (seq (addiu rs rs 0) is') *)

  | WFis_lw4 : 
    forall Phi G T b1 T1 rd rs is',
      (rd <> zero) ->
      (rlst_in rd tal_regs) ->
      regty_get G rs = Some (tmem b1 T1 true T) ->
      WFis Phi (regty_set G rd T) is' ->
      WFis Phi G (seq (lw rd 4 rs) is')

  | WFis_sw0 : 
    forall Phi G b T b2 T2 rd rs is',
      regty_get G rd = Some (tmem b T b2 T2) ->
      regty_get G rs = Some T ->
      WFis Phi (regty_set G rd (tmem true T b2 T2)) is' ->
      WFis Phi G (seq (sw rs 0 rd) is')

  | WFis_sw4 : 
    forall Phi G b T b1 T1 rd rs is',
      regty_get G rd = Some (tmem b1 T1 b T) ->
      regty_get G rs = Some T ->
      WFis Phi (regty_set G rd (tmem b1 T1 true T)) is' ->
      WFis Phi G (seq (sw rs 4 rd) is')

  | WFis_nulle :
    forall Phi G G' T rs is' lab,
      regty_get G rs = Some (tunion tnull T) ->
      chspec_in Phi lab (isp1 G') ->
      regty_sub (regty_set G rs tnull) G' ->
      WFis Phi (regty_set G rs T) is' ->
      WFis Phi G (beq rs zero lab is')
.


(** ******************************************************************* *)
(** soundness *)
(** ******************************************************************* *)
Theorem tal_sound_iseq : forall is P G,
  WFis P G is -> 
  iseq_ok (lift (interp (isp1 G)) P) is.
Proof.
  intros.
  induction H; rename Phi into P.

  (* j *)
  simpl; intros P' [rf hp] [Hlift [ht [hs [Hj [Hwf HR]]]]].
  exists (isp1 G'); split; auto.
  exists ht; exists hs; intuition.
  eapply WFst_regty_sub; eauto.

  (* jal *)
  simpl; intros P' [rf hp] [Hlift [ht [hs [Hj [Hwf HR]]]]].
  exists (isp1 G'); split; auto.
  exists ht; exists hs; intuition.
  apply WFst_regty_sub with (regty_set G ra (tcode G'')); trivial.
  destruct Hwf as [Phi [Hdh Hrf]].
  exists Phi; split; simpl; trivial.
  apply WFrf_rset; trivial.
  intro F; discriminate.
  right; right; right; left; reflexivity.
  apply WFw_code; auto.
  unfold WFwty.
  simpl.
  destruct H1; trivial.

  (* mov *)
  exists (lift (interp (isp1 (regty_set G rd T'))) P).
  split; trivial.
  intros P' [rf hp] [Hlift [ht [hs [Hj [Hwf HR]]]]]; simpl.
  split; auto.
  exists ht; exists hs; intuition.
  destruct Hwf as [Phi [Hdh Hrf]].
  exists Phi; split; simpl; trivial.
  apply WFrf_rset; trivial;
    destruct Hrf as [Hw [Hb [Hd [Hc Hz]]]].
  unfold rget; simplpm.
  apply (WFw_sub (Hw _ _ H1) H2).
  inversion_clear H2; unfold WFwty in * ; simpl in * ; trivial; splitb; trivial.
  hypreplace Hb (rec_bound_regty_elim Hc Hb).
  apply Hb with rs; trivial.
  hypreplace Hb (rec_bound_regty_elim Hc Hb).
  apply Hb with rs; trivial.
  
  apply gc_inv_rf_set_ok; trivial.

  (* lw *)
  exists (lift (interp (isp1 (regty_set G rd T))) P).
  split; trivial.
  intros P' [rf hp] [Hlift [ht [hs [Hj [Hwf HR]]]]]; simpl.

  destruct Hwf as [Phi [[Hdh Hdhwf] [Hwfw [Hr [Hb Hc]]]]].
  assert (Hrs := Hwfw _ _ H1).
  inversion_clear Hrs.
  destruct (Hdh _ _ _ H3) as [v1 [v2 [Hv1 [Hv2 [Hvc1 Hvc2]]]]].
  unfold hget; unfold rget; simpl in * .
  rewrite (heap_join_get_left Hv2 Hj).
  simpl.
  split; trivial.
  exists ht ; exists hs.
  repeat (split; trivial).
  exists Phi; split; simpl; trivial.
  split; trivial.
  destruct C2 as [b2 T2].
  assert (Hb2 : b2 = true);
    [inversion_clear H5; reflexivity | subst].
  injcty.
  apply WFrf_rset; trivial.
  repeat (split; trivial).
  inversion Hvc2; trivial.
  generalize (regty_bound_WFwty Hr H1).
  unfold WFwty; simpl.
  intros; elimbool; trivial.

  apply gc_inv_rf_set_ok; trivial.

  inversion_clear H4.

  (* sw 0 *)
  exists (lift (interp (isp1 (regty_set G rd (tmem true T b2 T2)))) P).
  split; trivial.
  intros P' [rf hp] [Hlift [ht [hs [Hj [Hwf HR]]]]]; simpl.

  destruct Hwf as [Phi [[Hdh Hdhwf] [Hwfw [Hr [Hb [Hc Hz]]]]]].
  assert (Hrs := Hwfw _ _ H).
  inversion_clear Hrs; [idtac | inversion H3].
  destruct (Hdh _ _ _ H2) as [v1 [v2 [Hv1 [Hv2 [Hvc1 Hvc2]]]]].
  unfold hget; unfold rget; simplpm; simpl in * ;
    rewrite (heap_join_get_left Hv1 Hj); split; trivial.

  exists (heap_set ht (rfile_get rf rd) (rfile_get rf rs)) ; exists hs.
  repeat split.
  eapply heap_join_set_left; eauto.

  destruct C1 as [b' T'].
  injcty.
  exists (dhspec_set Phi (rfile_get rf rd) (true, T) C2).
  split; trivial; simpl.
  apply heap_update_WFdh_o with b'; trivial.
  split; trivial.
  destruct b'; constructor.
  apply (Hwfw _ _ H0).
  
  destruct C2 as [b2' T2'].
  assert (A : WFrf P'
    (fun l' : nat =>
      if beq_nat l' (rfile_get rf rd) then Some ((true, T), (b2', T2')) else Phi l')
     rf G).
  apply heap_update_WFrf with b' b2'; trivial.
  repeat (split; trivial).
  destruct b'; constructor.
  constructor.

  clear Hwfw; destruct A as [Hwfw Hwfrf].
  split.
  intros r T'.
  comparer rd r.
  rewrite regty_set_get_eq; intros F;
    injection F; clear F; intros; subst.
  apply WFw_mem with (true, T) (b2', T2'); trivial.
  simplbeq;  reflexivity.
  constructor.
  rewrite regty_set_get_neq; auto.
  apply WFregty_set; trivial.
  intro F; subst.
  destruct Hz as [Z | [Z | Z]]; try discriminate; rewrite Z in H; discriminate.
  apply (regty_bound_WFwty Hr H).
  apply (regty_indom_intro H).

  apply gc_inv_heap_set_ok with v1; trivial.

  (* sw 4 *)
  exists (lift (interp (isp1 (regty_set G rd (tmem b1 T1 true T)))) P).
  split; trivial.
  intros P' [rf hp] [Hlift [ht [hs [Hj [Hwf HR]]]]]; simpl.

  destruct Hwf as [Phi [[Hdh Hdhwf] [Hwfw [Hr [Hb [Hc Hz]]]]]].
  assert (Hrs := Hwfw _ _ H).
  inversion_clear Hrs; [idtac | inversion H3].
  destruct (Hdh _ _ _ H2) as [v1 [v2 [Hv1 [Hv2 [Hvc1 Hvc2]]]]].
  unfold hget; unfold rget; simpl in * ; simplpm;
    rewrite (heap_join_get_left Hv2 Hj); split; auto.

  exists (heap_set ht (rfile_get rf rd + 4) (rfile_get rf rs)) ; exists hs.
  repeat split.
  eapply heap_join_set_left; eauto.

  destruct C2 as [b' T'].
  injcty.
  exists (dhspec_set Phi (rfile_get rf rd) C1 (true, T)).
  unfold rget.
  split; trivial; simpl.
  apply heap_update_WFdh_t with b'; trivial.
  split; trivial.
  destruct b'; constructor.
  apply (Hwfw _ _ H0).
  
  destruct C1 as [b1' T1'].
  assert (A : WFrf P'
    (fun l' : nat =>
      if beq_nat l' (rfile_get rf rd) then Some ((b1', T1'), (true, T)) else Phi l')
     rf G).
  apply heap_update_WFrf with b1' b'; trivial.
  repeat (split; trivial).
  destruct b'; constructor.
  destruct b'; constructor.

  clear Hwfw; destruct A as [Hwfw Hwfrf].
  split.
  intros r T'.
  comparer rd r.
  rewrite regty_set_get_eq; intros F;
    injection F; clear F; intros; subst.
  apply WFw_mem with (b1', T1') (true, T); trivial.
  simplbeq;  reflexivity.
  constructor.

  rewrite regty_set_get_neq; auto.

  apply WFregty_set; trivial.
  intro F; subst.
  destruct Hz as [Z | [Z | Z]]; try discriminate; rewrite Z in H; discriminate.
  apply (regty_bound_WFwty Hr H).
  apply (regty_indom_intro H).

  apply gc_inv_heap_set_ok with v2; trivial.

  (* null elim *)
  exists (lift (interp (isp1 (regty_set G rs T))) P).
  split; trivial.
  intros P' [rf hp] [Hlift [ht [hs [Hj [Hwf HR]]]]]; simpl.

  unfold step.
  simpl; dst; simplst.
  simplif.
  exists (isp1 G'); split; auto.
  exists ht; exists hs; intuition.
  apply WFst_regty_sub with (regty_set G rs tnull); trivial.
  destruct Hwf as [Phi [Hdh [Hrf Hrwf]]].
  exists Phi; split; trivial.
  split; trivial.
  intros.
  comparer rs r.
  rewrite regty_set_get_eq in H3.
  injection H3; intros; subst.
  unfold rget in R; simpl in * ;rewrite R.
  constructor.
  rewrite regty_set_get_neq in H3; auto.
  
  apply WFregty_set; trivial.
  intro F; subst.
  destruct Hrwf as [_ [_ [_ Hz]]].
  destruct Hz as [Z | [Z | Z]]; try discriminate; rewrite Z in H; discriminate.
  constructor.
  apply (regty_indom_intro H).

  split; trivial.
  exists ht; exists hs; intuition.
  destruct Hwf as [Phi [Hdh [Hrf Hrwf]]].
  exists Phi; split; trivial.
  split; trivial.
  intros.
  comparer rs r.
  rewrite regty_set_get_eq in H3.
  injection H3; intros; subst.
  hypreplace H (Hrf _ _ H).
  inversion_clear H.
  inversion H5.
  subst.
  elimtype False; apply R.
  unfold rget.
  apply (WFw_canon_tnull H4).
  subst; trivial.
  rewrite regty_set_get_neq in H3; auto.

  apply WFregty_set; trivial.
  destruct Hrwf as [Hr [Hb [Hc Hz]]].
  intro F; subst.
  destruct Hz as [Z | [Z | Z]]; try discriminate; rewrite Z in H; discriminate.
  destruct Hrwf as [Hr [Hb [Hc Hz]]].
  hypreplace H (regty_bound_WFwty Hr H).
  inversion H; trivial.
  apply (regty_indom_intro H).
Qed.

(* (\** ******************************************************************* *\) *)
(* (\** stand alone soundness *\) *)
(* (\** ******************************************************************* *\) *)
(* Lemma WFprog_is_inv : forall Psi p G, *)
(*   WFprog Psi G p -> exists p', exists G', step p = Some p' /\ WFprog Psi G' p'. *)
(* Proof. *)
(*   intros Psi [[ch st] is] G. *)
(*   unfold WFprog; dst. *)
(*   intros [Hcs [Hch [Hst His]]]. *)
(*   inversion_clear His. *)

(*   (\* jd *\) *)
(*   hypreplace H (Hch _ _ H). *)
(*   simpl in H. *)
(*   destruct (ch_lab_dec ch lab) as [[is' His'] | F]. *)
(*   rewrite His' in H. *)
(*   exists (ch, st, is'); exists G'; dst. *)
(*   repeat (split; trivial). *)
(*   unfold step. *)
(*   dst. *)
(*   simpl. *)
(*   rewrite His'. *)
(*   reflexivity. *)

(*   apply (WFst_regty_sub Hst H0). *)
(*   rewrite F in H; tauto. *)
  
(*   (\* jal *\) *)
(*   hypreplace H (Hch _ _ H). *)
(*   simpl in H. *)
(*   destruct (ch_lab_dec ch lab) as [[is' His'] | F]. *)
(*   rewrite His' in H. *)
(*   exists (ch, (rset st ra fret), is'); exists G'; dst. *)
(*   repeat (split; trivial). *)
(*   unfold step. *)
(*   dst. *)
(*   simpl. *)
(*   rewrite His'. *)
(*   reflexivity. *)

(*   apply WFst_regty_sub with (regty_set G ra (tcode G'')); trivial. *)
(*   destruct Hst as [Phi [Hdh Hrf]]. *)
(*   exists Phi; split; simpl; trivial. *)
(*   apply WFrf_rset; trivial. *)
(*   intro F; discriminate. *)
(*   right; right; right; left; reflexivity. *)
(*   apply WFw_code with G''; auto. *)
(*   constructor. *)
(*   firstorder. *)
(*   apply (Hcs _ _ H0). *)
(*   unfold WFwty. *)
(*   simpl. *)
(*   destruct (Hcs _ _ H0); trivial. *)
(*   rewrite F in H; tauto. *)

(*   (\* mov *\) *)
(*   exists (ch, (rset st rd (rget st rs)), is'); exists (regty_set G rd T); dst. *)
(*   repeat (split; trivial). *)
(*   unfold step. *)
(*   simpl. *)
(*   simplpm. *)
(*   reflexivity. *)

(*   destruct Hst as [Phi [Hdh Hrf]]. *)
(*   exists Phi; split; simpl; trivial. *)
(*   apply WFrf_rset; trivial; *)
(*     destruct Hrf as [Hw [Hb [Hd [Hc Hz]]]]. *)
(*   unfold rget. *)
(*   apply Hw; trivial. *)
(*   unfold WFwty. *)
(*   hypreplace Hb (rec_bound_regty_elim Hc Hb). *)
(*   apply Hb with rs; trivial. *)

(*   (\* subt *\) *)
(*   exists (ch, (rset st rs (rget st rs)), is'); exists (regty_set G rs T'); dst. *)
(*   repeat (split; trivial). *)
(*   unfold step. *)
(*   simpl; simplpm; *)
(*     reflexivity. *)

(*   destruct Hst as [Phi [Hdh Hrf]]. *)
(*   exists Phi; split; simpl; trivial. *)
(*   apply WFrf_rset; trivial; *)
(*     destruct Hrf as [Hw [Hb [Hd Hc]]]. *)
(*   unfold rget. *)
(*   apply WFw_sub with T; auto. *)
(*   apply WFwty_wty_sub with T; trivial. *)
(*   apply regty_bound_WFwty with G rs; trivial. *)

(*   (\* lw *\) *)
(*   destruct Hst as [Phi [[Hdh Hdhwf] [Hwfw [Hr [Hb Hc]]]]]. *)
(*   assert (Hrs := Hwfw _ _ H1). *)
(*   inversion_clear Hrs. *)
(*   destruct (Hdh _ _ _ H3) as [v1 [v2 [Hv1 [Hv2 [Hvc1 Hvc2]]]]]. *)
  
(*   exists (ch, (rset st rd v2), is'); exists (regty_set G rd T); dst. *)
(*   repeat (split; trivial). *)
(*   unfold step. *)
(*   simpl; unfold hget; unfold rget; dst;  *)
(*     rewrite Hv2; reflexivity.  *)
  
(*   exists Phi; split; simpl; trivial. *)
(*   split; trivial. *)
(*   destruct C2 as [b2 T2]. *)
(*   assert (Hb2 : b2 = true); *)
(*     [inversion_clear H5; reflexivity | subst]. *)
(*   injcty. *)
(*   apply WFrf_rset; trivial. *)
(*   repeat (split; trivial). *)
(*   inversion Hvc2; trivial. *)
(*   generalize (regty_bound_WFwty Hr H1). *)
(*   unfold WFwty; simpl. *)
(*   intros; elimbool; trivial. *)
(*   inversion H4. *)

(*   (\* sw 0 *\) *)
(*   destruct Hst as [Phi [[Hdh Hdhwf] [Hwfw [Hr [Hb [Hc Hz]]]]]]. *)
(*   assert (Hrs := Hwfw _ _ H). *)
(*   inversion_clear Hrs; [idtac | inversion H3]. *)
(*   destruct (Hdh _ _ _ H2) as [v1 [v2 [Hv1 [Hv2 [Hvc1 Hvc2]]]]]. *)
  
(*   exists (ch, (hset st (rget st rd) (rget st rs)), is'). *)
(*   exists (regty_set G rd (tmem true T b2 T2)); dst. *)
(*   repeat (split; trivial). *)
(*   unfold step. *)
(*   simpl; unfold hget; unfold rget; dst; simplpm; *)
(*     rewrite Hv1; reflexivity. *)

(*   destruct C1 as [b' T']. *)
(*   injcty. *)
(*   exists (dhspec_set Phi (rfile_get (rf_of st) rd) (true, T) C2). *)
(*   split; trivial; simpl. *)
(*   apply heap_update_WFdh_o with b'; trivial. *)
(*   split; trivial. *)
(*   destruct b'; constructor. *)
(*   apply (Hwfw _ _ H0). *)
  
(*   destruct C2 as [b2' T2']. *)
(*   assert (A : WFrf Psi *)
(*     (fun l' : nat => *)
(*       if beq_nat l' (rfile_get (rf_of st) rd) then Some ((true, T), (b2', T2')) else Phi l') *)
(*      (rf_of st) G). *)
(*   apply heap_update_WFrf with b' b2'; trivial. *)
(*   repeat (split; trivial). *)
(*   destruct b'; constructor. *)
(*   constructor. *)

(*   clear Hwfw; destruct A as [Hwfw Hwfrf]. *)
(*   split. *)
(*   intros r T'. *)
(*   comparer rd r. *)
(*   rewrite regty_set_get_eq; intros F;  *)
(*     injection F; clear F; intros; subst. *)
(*   apply WFw_mem with (true, T) (b2', T2'); trivial. *)
(*   simplbeq;  reflexivity. *)
(*   constructor. *)
(*   rewrite regty_set_get_neq; auto. *)
(*   apply WFregty_set; trivial. *)
(*   intro F; subst. *)
(*   destruct Hz as [Z | Z]; rewrite Z in H; discriminate. *)
(*   apply (regty_bound_WFwty Hr H). *)
(*   apply (regty_indom_intro H). *)

(*   (\* sw 4 *\) *)
(*   destruct Hst as [Phi [[Hdh Hdhwf] [Hwfw [Hr [Hb [Hc Hz]]]]]]. *)
(*   assert (Hrs := Hwfw _ _ H). *)
(*   inversion_clear Hrs; [idtac | inversion H3]. *)
(*   destruct (Hdh _ _ _ H2) as [v1 [v2 [Hv1 [Hv2 [Hvc1 Hvc2]]]]]. *)
  
(*   exists (ch, (hset st (rget st rd + 4) (rget st rs)), is'). *)
(*   exists (regty_set G rd (tmem b1 T1 true T)); dst. *)
(*   repeat (split; trivial). *)
(*   unfold step. *)
(*   simpl; unfold hget; unfold rget; dst; simplpm; *)
(*     rewrite Hv2; reflexivity. *)

(*   destruct C2 as [b' T']. *)
(*   injcty. *)
(*   exists (dhspec_set Phi (rfile_get (rf_of st) rd) C1 (true, T)). *)
(*   unfold rget. *)
(*   split; trivial; simpl. *)
(*   apply heap_update_WFdh_t with b'; trivial. *)
(*   split; trivial. *)
(*   destruct b'; constructor. *)
(*   apply (Hwfw _ _ H0). *)
  
(*   destruct C1 as [b1' T1']. *)
(*   assert (A : WFrf Psi *)
(*     (fun l' : nat => *)
(*       if beq_nat l' (rfile_get (rf_of st) rd) then Some ((b1', T1'), (true, T)) else Phi l') *)
(*      (rf_of st) G). *)
(*   apply heap_update_WFrf with b1' b'; trivial. *)
(*   repeat (split; trivial). *)
(*   destruct b'; constructor. *)
(*   destruct b'; constructor. *)

(*   clear Hwfw; destruct A as [Hwfw Hwfrf]. *)
(*   split. *)
(*   intros r T'. *)
(*   comparer rd r. *)
(*   rewrite regty_set_get_eq; intros F;  *)
(*     injection F; clear F; intros; subst. *)
(*   apply WFw_mem with (b1', T1') (true, T); trivial. *)
(*   simplbeq;  reflexivity. *)
(*   constructor. *)

(*   rewrite regty_set_get_neq; auto. *)

(*   apply WFregty_set; trivial. *)
(*   intro F; subst. *)
(*   destruct Hz as [Z | Z]; rewrite Z in H; discriminate. *)
(*   apply (regty_bound_WFwty Hr H). *)
(*   apply (regty_indom_intro H). *)

(*   (\* null elim *\) *)
(*   unfold step. *)
(*   simpl; dst; simplst. *)
(*   simplif. *)
(*   hypreplace H0 (Hch _ _ H0). *)
(*   simpl in H0. *)
(*   destruct (ch_lab_dec ch lab) as [[is'' His''] | F]. *)
(*   rewrite His'' in H0. *)
(*   exists (ch, st, is''); exists G'; dst. *)
(*   repeat (split; trivial). *)
(*   rewrite His''. *)
(*   reflexivity. *)

(*   apply WFst_regty_sub with (regty_set G rs tnull); trivial. *)
(*   destruct Hst as [Phi [Hdh [Hrf Hrwf]]]. *)
(*   exists Phi; split; trivial. *)
(*   split; trivial. *)
(*   intros. *)
(*   comparer rs r. *)
(*   rewrite regty_set_get_eq in H3. *)
(*   injection H3; intros; subst. *)
(*   unfold rget in R; rewrite R. *)
(*   constructor. *)
(*   rewrite regty_set_get_neq in H3; auto. *)
  
(*   apply WFregty_set; trivial. *)
(*   intro F; subst. *)
(*   destruct Hrwf as [_ [_ [_ Hz]]]. *)
(*   destruct Hz as [Z | Z]; rewrite Z in H; discriminate. *)
(*   constructor. *)
(*   apply (regty_indom_intro H). *)

(*   rewrite F in H0; inversion H0. *)

(*   exists (ch, st, is'); exists (regty_set G rs T); dst. *)
(*   repeat (split; trivial). *)
(*   destruct Hst as [Phi [Hdh [Hrf Hrwf]]]. *)
(*   exists Phi; split; trivial. *)
(*   split; trivial. *)
(*   intros. *)
(*   comparer rs r. *)
(*   rewrite regty_set_get_eq in H3. *)
(*   injection H3; intros; subst. *)
(*   hypreplace H (Hrf _ _ H). *)
(*   inversion_clear H. *)
(*   inversion H5. *)
(*   subst. *)
(*   elimtype False; apply R. *)
(*   unfold rget. *)
(*   apply (WFw_canon_tnull H4). *)
(*   subst; trivial. *)
(*   rewrite regty_set_get_neq in H3; auto. *)

(*   apply WFregty_set; trivial. *)
(*   destruct Hrwf as [Hr [Hb [Hc Hz]]]. *)
(*   intro F; subst. *)
(*   destruct Hz as [Z | Z]; rewrite Z in H; discriminate. *)
(*   destruct Hrwf as [Hr [Hb [Hc Hz]]]. *)
(*   hypreplace H (regty_bound_WFwty Hr H). *)
(*   inversion H; trivial. *)
(*   apply (regty_indom_intro H). *)
(* Qed. *)

(* TODO: *)

(* 1. limit reg to s1 ~ s4 *)
(* 2. well-formed type? no free var. *)



(*   destruct Frf as [Hbound [Hsubset Hcanon]]. *)
(* Lemma refty_sub_rec_bound : forall G G', *)
(*   regty_canon G -> rec_bound_regty 0 G = true ->  *)
(*   regty_sub G G' -> rec_bound_regty 0 G' = true. *)
(* Proof. *)
(*   intros. *)
(*   assert (Hsub := regty_sub_dom_sub H1). *)
(*   apply rec_bound_regty_intro. *)
(* Lemma  *)
(*   reflexivity. *)
  
(*   reflexivity. *)


(* Lemma regty_sub_nil_nil : forall G,  *)
(*   regty_sub regty_nil G -> G = regty_nil. *)
(* Proof. *)
(*   induction G; intros. *)
(*   reflexivity. *)
(*   generalize (H r); clear H. *)
(*   simpl; solvereg; intro. *)
(*   hypreplace H (H (or_introl _ (refl_equal r))). *)
(*   discriminate. *)
(* Qed. *)
(*   hypreplace H0 (regty_sub_nil_nil H0). *)
(*   rewrite H0; reflexivity. *)
(*   intros. *)
(*   apply IHG. *)
(*   simpl in H. *)
(*   elimbool; trivial. *)
(*   apply regty_sub_trans with (regty_cons r w G); auto. *)
(*   apply regty_sub_dom_sub. *)



(* Lemma regty_sub_wfty : forall G' G, *)
(*   rec_bound_regty 0 G = true -> regty_sub G G' ->  *)
(*   rec_bound_regty 0 G' = true. *)
(* Proof. *)
(*   induction G'; simpl; intros. *)
(*   reflexivity. *)
(*   splitb. *)
(*   destruct (H0 r). *)
(*   simpl in H2. *)
(*   rewrite (beq_reg_refl r) in H2. *)
(*   Lemma reg_indom_wf : forall r G, *)
(*     rec_bound_regty 0 G = true -> regty_ *)
(*   simpl. *)

(*   simpl. *)


(* regty_sub *)
(* Lemma regty_sub_trans : forall G1 G2 G3, *)
(*   regty_sub G1 G2 -> regty_sub G2 G3 -> regty_sub G1 G3. *)
(* Proof. *)
(*   intros; destruct H; destruct H0. *)
(*   split; trivial. *)
(*   intros. *)
(*   hypreplace H (H r); hypreplace H0 (H0 _ H3). *)
(*   rewrite H0; apply H. *)
(*   destruct (regty_indom_elim H3). *)
(*   rewrite H4 in H0. *)
(*   apply (regty_indom_intro (sym_eq H0)). *)
(* Qed. *)

(* Lemma regty_sub_dom_sub : forall G' G, *)
(*   regty_sub G G' -> rlst_subset (regty_dom G') (regty_dom G). *)
(* Proof. *)
(*   red; intros. *)
(*   destruct H. *)
(*   hypreplace H (H _ H0). *)
(*   destruct (regty_indom_elim H0). *)
(*   apply regty_indom_intro with x. *)
(*   rewrite <- H; apply H2. *)
(* Qed. *)
